@@ -54,7 +54,10 @@ class Addon:
         except TypeError:
             self.obj = self.module.init()
         if self.obj:
-            self.info = self.obj.info()
+            try:
+                self.info = self.obj.info(**kw)
+            except TypeError:
+                self.info = self.obj.info()
 
     def finished(self):
         if hasattr(self.obj, "finished"):
@@ -65,6 +68,43 @@ class Addon:
         if func is not None:
             func(**kw)
 
+class Plugins(list):
+    def __init__(self, **kw):
+        """
+        Read the plugins, in sorted order.
+
+        Pass all the keyword args to the plugin init
+        """
+        if "langs" not in kw:
+            kw["langs"] = self.scan_available_languages()
+        progress = kw.get("progress", None)
+
+        self.disabled = []
+        for fname in sorted(os.listdir(axi.PLUGINDIR)):
+            # Skip non-pythons, hidden files and python sources starting with '_'
+            if fname[0] in ['.', '_'] or not fname.endswith(".py"): continue
+            fullname = os.path.join(axi.PLUGINDIR, fname)
+            if not os.path.isfile(fullname): continue
+            if progress: progress.verbose("Reading plugin %s." % fullname)
+            addon = Addon(fullname, **kw)
+            if addon.obj != None:
+                self.append(addon)
+
+    def scan_available_languages(self):
+        # Languages we index
+        langs = set()
+
+        # Look for files like: ftp.uk.debian.org_debian_dists_sid_main_i18n_Translation-it
+        # And extract the language code at the end
+        tfile = re.compile(r"_i18n_Translation-([^-]+)$")
+        for f in os.listdir(APTLISTDIR):
+            mo = tfile.search(f)
+            if not mo: continue
+            langs.add(urllib.unquote(mo.group(1)))
+
+        return langs
+
+
 class Progress:
     """
     Normal progress report to stdout
@@ -173,33 +213,33 @@ class ClientProgress:
             try:
                 args = pickle.loads(msg)
             except EOFError:
-                progress.error("The other update has stopped")
+                self.progress.error("The other update has stopped")
                 return
             action = args[0]
             args = args[1:]
             if action == "begin":
-                progress.begin(*args)
+                self.progress.begin(*args)
                 hasBegun = True
             elif action == "progress":
                 if not hasBegun:
-                    progress.begin(args[0])
+                    self.progress.begin(args[0])
                     hasBegun = True
-                progress.progress(*args[1:])
+                self.progress.progress(*args[1:])
             elif action == "end":
                 if not hasBegun:
-                    progress.begin(args[0])
+                    self.progress.begin(args[0])
                     hasBegun = True
-                progress.end(*args[1:])
+                self.progress.end(*args[1:])
             elif action == "verbose":
-                progress.verbose(*args)
+                self.progress.verbose(*args)
             elif action == "notice":
-                progress.notice(*args)
+                self.progress.notice(*args)
             elif action == "error":
-                progress.error(*args)
+                self.progress.error(*args)
             elif action == "alldone":
                 break
             else:
-                progress.error("unknown action '%s' from other update-apt-xapian-index.  Arguments: '%s'" % (action, ", ".join(map(repr, args))))
+                self.progress.error("unknown action '%s' from other update-apt-xapian-index.  Arguments: '%s'" % (action, ", ".join(map(repr, args))))
 
 
 class ServerSenderProgress:
@@ -314,16 +354,10 @@ class Indexer(object):
         self.ds_timestamp = 0
         # Apt cache instantiated on demand
         self.apt_cache = None
-        # Languages we index
-        self.langs = set()
-
-        # Look for files like: ftp.uk.debian.org_debian_dists_sid_main_i18n_Translation-it
-        # And extract the language code at the end
-        tfile = re.compile(r"_i18n_Translation-([^-]+)$")
-        for f in os.listdir(APTLISTDIR):
-            mo = tfile.search(f)
-            if not mo: continue
-            self.langs.add(urllib.unquote(mo.group(1)))
+        # Loaded plugins
+        self.plugins = None
+        # OS file handle for the lock file
+        self.lockfd = None
 
         # Ensure the database and cache directories exist
         self.ensure_dir_exists(axi.XAPIANDBPATH)
@@ -364,6 +398,7 @@ class Indexer(object):
             ##warnings.simplefilter('ignore', FutureWarning)
             #warnings.filterwarnings("ignore","apt API not stable yet")
             import apt
+            import apt_pkg
             #warnings.resetwarnings()
 
             if self.quietapt:
@@ -372,13 +407,14 @@ class Indexer(object):
                         pass
                     def done(self):
                         pass
-                    def update(self,percent):
+                    def update(self,percent=None):
                         pass
                 aptprogress = AptSilentProgress()
             else:
                 aptprogress = None
 
             # memonly=True: force apt to not write a pkgcache.bin
+            apt_pkg.init_config()
             self.apt_cache = apt.Cache(memonly=True, progress=aptprogress)
         return self.apt_cache
 
@@ -391,10 +427,9 @@ class Indexer(object):
           False if another indexer is running
         """
         # Lock the session so that we prevent concurrent updates
-        lockfd = os.open(axi.XAPIANDBLOCK, os.O_RDWR | os.O_CREAT)
-        lockpyfd = os.fdopen(lockfd)
+        self.lockfd = os.open(axi.XAPIANDBLOCK, os.O_RDWR | os.O_CREAT)
         try:
-            fcntl.lockf(lockpyfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
+            fcntl.lockf(self.lockfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
             # Wrap the current progress with the server sender
             self.progress = ServerProgress(self.progress)
             return True
@@ -410,28 +445,11 @@ class Indexer(object):
 
         Return when the other indexer has finished.
         """
-        progress.notice("Another update is already running: showing its progress.")
+        self.progress.notice("Another update is already running: showing its progress.")
         childProgress = ClientProgress(self.progress)
         childProgress.loop()
 
-    def readPlugins(self, **kw):
-        """
-        Read the addons, in sorted order.
-
-        Pass all the keyword args to the plugin init
-        """
-        self.addons = []
-        for fname in sorted(os.listdir(axi.PLUGINDIR)):
-            # Skip non-pythons, hidden files and python sources starting with '_'
-            if fname[0] in ['.', '_'] or not fname.endswith(".py"): continue
-            fullname = os.path.join(axi.PLUGINDIR, fname)
-            if not os.path.isfile(fullname): continue
-            self.progress.verbose("Reading plugin %s." % fullname)
-            addon = Addon(fullname, **kw)
-            if addon.obj != None:
-                self.addons.append(addon)
-
-    def setupIndexing(self, force=False):
+    def setupIndexing(self, force=False, system=True):
         """
         Setup indexing: read plugins, check timestamps...
 
@@ -444,16 +462,16 @@ class Indexer(object):
         # Read values database
         #values = readValueDB(VALUESCONF, progress)
 
-        # Read the addons, in sorted order
-        self.readPlugins(langs=self.langs, progress=self.progress)
+        # Read the plugins, in sorted order
+        self.plugins = Plugins(progress=self.progress, system=system)
 
         # Ensure that we have something to do
-        if len(self.addons) == 0:
+        if len(self.plugins) == 0:
             self.progress.notice("No indexing plugins found in %s" % axi.PLUGINDIR)
             return False
 
         # Get the most recent modification timestamp of the data sources
-        self.ds_timestamp = max([x.info['timestamp'] for x in self.addons])
+        self.ds_timestamp = max([x.info['timestamp'] for x in self.plugins])
 
         # Get the timestamp of the last database update
         try:
@@ -470,7 +488,7 @@ class Indexer(object):
             self.progress.verbose("Most recent update for: %s." % time.ctime(cur_timestamp))
 
         # See if we need an update
-        if int(self.ds_timestamp+.5) <= int(cur_timestamp+0.5):
+        if cur_timestamp != 0 and int(self.ds_timestamp+.5) <= int(cur_timestamp+0.5):
             if force:
                 self.progress.notice("The index %s is up to date, but rebuilding anyway as requested." % axi.XAPIANDBPATH)
             else:
@@ -482,17 +500,17 @@ class Indexer(object):
         # Read existing value database to keep ids stable in a system
         self.values, self.values_desc = axi.readValueDB(quiet=True)
         values_seq = max(self.values.values()) + 1
-        for addon in self.addons:
+        for addon in self.plugins:
             for v in addon.info.get("values", []):
                 if v['name'] in self.values: continue
                 self.values[v['name']] = values_seq
                 values_seq += 1
                 self.values_desc[v['name']] = v['desc']
 
-        # Tell the addons to do the long initialisation bits
+        # Tell the plugins to do the long initialisation bits
         self.progress.verbose("Initializing plugins.")
-        for addon in self.addons:
-            addon.obj.init(dict(values = self.values), self.progress)
+        for addon in self.plugins:
+            addon.obj.init(dict(values=self.values), self.progress)
 
         return True
 
@@ -509,7 +527,7 @@ class Indexer(object):
         # document by exact package name match
         document.add_term("XP"+pkg.name)
         # Have all the various plugins index their things
-        for addon in self.addons:
+        for addon in self.plugins:
             addon.obj.index(document, pkg)
         return document
 
@@ -527,7 +545,7 @@ class Indexer(object):
         # document by exact package name match
         document.add_term("XP"+pkg["Package"])
         # Have all the various plugins index their things
-        for addon in self.addons:
+        for addon in self.plugins:
             addon.obj.indexDeb822(document, pkg)
         return document
 
@@ -544,21 +562,26 @@ class Indexer(object):
             if idx % 200 == 0: self.progress.progress(100*idx/count)
             yield self.get_document_from_apt(pkg)
 
-    def gen_documents_deb822(self, fname):
+    def gen_documents_deb822(self, fnames):
         try:
             from debian import deb822
         except ImportError:
             from debian_bundle import deb822
-        infd = open(fname)
-        # Get file size to compute progress
-        total = os.fstat(infd.fileno())[6]
-        for idx, pkg in enumerate(deb822.Deb822.iter_paragraphs(infd)):
-            # Print approximate progress by checking the current read position
-            # against the file size
-            if total > 0 and idx % 200 == 0:
-                cur = infd.tell()
-                self.progress.progress(100*cur/total)
-            yield self.get_document_from_deb822(pkg)
+        seen = set()
+        for fname in fnames:
+            infd = open(fname)
+            # Get file size to compute progress
+            total = os.fstat(infd.fileno())[6]
+            for idx, pkg in enumerate(deb822.Deb822.iter_paragraphs(infd)):
+                name = pkg["Package"]
+                if name in seen: continue
+                seen.add(name)
+                # Print approximate progress by checking the current read position
+                # against the file size
+                if total > 0 and idx % 200 == 0:
+                    cur = infd.tell()
+                    self.progress.progress(100*cur/total)
+                yield self.get_document_from_deb822(pkg)
 
     def compareCacheToDb(self, cache, db):
         """
@@ -609,7 +632,7 @@ class Indexer(object):
                                                     len(obsolete)))
 
         self.progress.begin("Updating Xapian index")
-        for a in self.addons: a.send_extra_info(db=db, aptcache=cache)
+        for a in self.plugins: a.send_extra_info(db=db, aptcache=cache)
         for idx, pkg in enumerate(cache):
             if idx % 1000 == 0: self.progress.progress(100*idx/count)
             if not pkg.candidate:
@@ -628,7 +651,7 @@ class Indexer(object):
             db.delete_document(docid)
 
         # finished
-        for a in self.addons:
+        for a in self.plugins:
             a.finished()
         db.flush()
         self.progress.end()
@@ -644,11 +667,12 @@ class Indexer(object):
         # Update the index timestamp
         if not os.path.exists(axi.XAPIANDBSTAMP):
             open(axi.XAPIANDBSTAMP, "w").close()
-        os.utime(axi.XAPIANDBSTAMP, (self.ds_timestamp, self.ds_timestamp))
+        if self.ds_timestamp > 0:
+            os.utime(axi.XAPIANDBSTAMP, (self.ds_timestamp, self.ds_timestamp))
 
     def buildIndex(self, pathname, documents, addoninfo={}):
         """
-        Create a new Xapian index with the content provided by the addons
+        Create a new Xapian index with the content provided by the plugins
         """
         self.progress.begin("Rebuilding Xapian index")
 
@@ -657,20 +681,20 @@ class Indexer(object):
         # It seems to be faster without transactions, at the moment
         #db.begin_transaction(False)
 
-        for a in self.addons: a.send_extra_info(db=db)
+        for a in self.plugins: a.send_extra_info(db=db)
 
         # Add all generated documents to the index
         for doc in documents:
             db.add_document(doc)
 
         #db.commit_transaction();
-        for a in self.addons: 
+        for a in self.plugins: 
             a.finished()
         db.flush()
         self.progress.end()
 
-    def rebuild(self, pkgfile=None):
-        # Create a new Xapian index with the content provided by the addons
+    def rebuild(self, pkgfiles=None):
+        # Create a new Xapian index with the content provided by the plugins
         # Xapian takes care of preventing concurrent updates and removing the old
         # database if it's left over by a previous crashed update
 
@@ -680,10 +704,10 @@ class Indexer(object):
             dbdir = os.path.join(axi.XAPIANCACHEPATH, tmpidxfname)
             if not os.path.exists(dbdir): break;
 
-        if pkgfile:
-            generator = self.gen_documents_deb822(pkgfile)
+        if pkgfiles:
+            generator = self.gen_documents_deb822(pkgfiles)
         else:
-            for a in self.addons: a.send_extra_info(aptcache=self.aptcache())
+            for a in self.plugins: a.send_extra_info(aptcache=self.aptcache())
             generator = self.gen_documents_apt()
         self.buildIndex(dbdir, generator)
 
@@ -692,7 +716,7 @@ class Indexer(object):
 
         #os.symlink(tmpidxfname, axi.XAPIANDBPATH + "/index.tmp")
         out = open(axi.XAPIANINDEX + ".tmp", "w")
-        print >>out, "auto", dbdir
+        print >>out, "auto", os.path.abspath(dbdir)
         out.close()
         os.rename(axi.XAPIANINDEX + ".tmp", axi.XAPIANINDEX)
 
@@ -713,11 +737,59 @@ class Indexer(object):
         # Commit the changes and update the last update timestamp
         if not os.path.exists(axi.XAPIANDBSTAMP):
             open(axi.XAPIANDBSTAMP, "w").close()
-        os.utime(axi.XAPIANDBSTAMP, (self.ds_timestamp, self.ds_timestamp))
+        if self.ds_timestamp > 0:
+            os.utime(axi.XAPIANDBSTAMP, (self.ds_timestamp, self.ds_timestamp))
 
         self.writeValues()
+        self.writePrefixes()
         self.writeDoc()
 
+    def writePrefixes(self, pathname=axi.XAPIANDBPREFIXES):
+        """
+        Write the prefix information on the given file
+        """
+        self.progress.verbose("Writing prefix information to %s." % pathname)
+        out = open(pathname+".tmp", "w")
+
+        print >>out, textwrap.dedent("""
+        # This file contains the information about keyword prefixes used in the
+        # APT Xapian index.
+        #
+        # Xapian allows to prefix some terms so they can be told apart from
+        # normal keywords: this is used for example with Debtags tags, and
+        # stemmed forms.
+        #
+        # This file lists terms with their index prefix, their queryparser
+        # prefix, whether queryparser should treat it as boolean or
+        # probabilistic and a short description.
+        """).lstrip()
+
+        # Aggregate and normalise prefix information from all the plugins
+        prefixes = dict()
+        for addon in self.plugins:
+            for p in addon.info.get("prefixes", []):
+                idx = p.get("idx", None)
+                if idx is None: continue
+                qp = p.get("qp", None)
+                type = p.get("type", None)
+                desc = p.get("desc", None)
+                # TODO: warn of inconsistencies (plugins that disagree on qp or type)
+                old = prefixes.setdefault(idx, dict())
+                if qp: old.setdefault("qp", qp)
+                if type: old.setdefault("type", type)
+                if desc: old.setdefault("desc", desc)
+
+        for name, info in sorted(prefixes.iteritems(), key=lambda x: x[0]):
+            print >>out, "%s\t%s\t%s\t# %s" % (
+                name,
+                info.get("qp", "-"),
+                info.get("type", "-"),
+                info.get("desc", "(description is missing)"))
+
+        out.close()
+        # Atomic update of the documentation
+        os.rename(pathname+".tmp", pathname)
+
     def writeValues(self, pathname=axi.XAPIANDBVALUES):
         """
         Write the value information on the given file
@@ -754,7 +826,7 @@ class Indexer(object):
         self.progress.verbose("Writing documentation to %s." % pathname)
         # Collect the documentation
         docinfo = []
-        for addon in self.addons:
+        for addon in self.plugins:
             try:
                 doc = addon.obj.doc()
                 if doc != None:
